# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
import six

"""Contains classes and functions that a SAML2.0 Service Provider (SP) may use
to conclude its tasks.
"""
from saml2.request import LogoutRequest
import saml2

from saml2 import saml, SAMLError
from saml2 import BINDING_HTTP_REDIRECT
from saml2 import BINDING_HTTP_POST
from saml2 import BINDING_SOAP

from saml2.ident import decode, code
from saml2.httpbase import HTTPError
from saml2.s_utils import sid
from saml2.s_utils import status_message_factory
from saml2.s_utils import success_status_factory
from saml2.samlp import STATUS_REQUEST_DENIED
from saml2.samlp import STATUS_UNKNOWN_PRINCIPAL
from saml2.time_util import not_on_or_after
from saml2.saml import AssertionIDRef
from saml2.client_base import Base
from saml2.client_base import SignOnError
from saml2.client_base import LogoutError
from saml2.client_base import NoServiceDefined
from saml2.mdstore import locations

import logging

logger = logging.getLogger(__name__)


class Saml2Client(Base):
    """ The basic pySAML2 service provider class """

    def prepare_for_authenticate(
        self,
        entityid=None,
        relay_state="",
        binding=saml2.BINDING_HTTP_REDIRECT,
        vorg="",
        nameid_format=None,
        scoping=None,
        consent=None, extensions=None,
        sign=None,
        sigalg=None,
        digest_alg=None,
        response_binding=saml2.BINDING_HTTP_POST,
        **kwargs,
    ):
        """ Makes all necessary preparations for an authentication request.

        :param entityid: The entity ID of the IdP to send the request to
        :param relay_state: To where the user should be returned after
            successfull log in.
        :param binding: Which binding to use for sending the request
        :param vorg: The entity_id of the virtual organization I'm a member of
        :param nameid_format:
        :param scoping: For which IdPs this query are aimed.
        :param consent: Whether the principal have given her consent
        :param extensions: Possible extensions
        :param sign: Whether the request should be signed or not.
        :param response_binding: Which binding to use for receiving the response
        :param kwargs: Extra key word arguments
        :return: session id and AuthnRequest info
        """

        reqid, negotiated_binding, info = self.prepare_for_negotiated_authenticate(
            entityid=entityid,
            relay_state=relay_state,
            binding=binding,
            vorg=vorg,
            nameid_format=nameid_format,
            scoping=scoping,
            consent=consent,
            extensions=extensions,
            sign=sign,
            sigalg=sigalg,
            digest_alg=digest_alg,
            response_binding=response_binding,
            **kwargs,
        )

        if negotiated_binding != binding:
            raise ValueError(
                "Negotiated binding '{}' does not match binding to use '{}'".format(
                    negotiated_binding, binding
                )
            )

        return reqid, info

    def prepare_for_negotiated_authenticate(
        self,
        entityid=None,
        relay_state="",
        binding=None,
        vorg="",
        nameid_format=None,
        scoping=None,
        consent=None,
        extensions=None,
        sign=None,
        response_binding=saml2.BINDING_HTTP_POST,
        sigalg=None,
        digest_alg=None,
        **kwargs,
    ):
        """ Makes all necessary preparations for an authentication request
        that negotiates which binding to use for authentication.

        :param entityid: The entity ID of the IdP to send the request to
        :param relay_state: To where the user should be returned after
            successfull log in.
        :param binding: Which binding to use for sending the request
        :param vorg: The entity_id of the virtual organization I'm a member of
        :param nameid_format:
        :param scoping: For which IdPs this query are aimed.
        :param consent: Whether the principal have given her consent
        :param extensions: Possible extensions
        :param sign: Whether the request should be signed or not.
        :param response_binding: Which binding to use for receiving the response
        :param kwargs: Extra key word arguments
        :return: session id and AuthnRequest info
        """

        expected_binding = binding

        for binding in [BINDING_HTTP_REDIRECT, BINDING_HTTP_POST]:
            if expected_binding and binding != expected_binding:
                continue

            destination = self._sso_location(entityid, binding)
            logger.info("destination to provider: %s", destination)

            # XXX - sign_post will embed the signature to the xml doc
            # XXX   ^through self.create_authn_request(...)
            # XXX - sign_redirect will add the signature to the query params
            # XXX   ^through self.apply_binding(...)
            sign_post = False if binding == BINDING_HTTP_REDIRECT else sign
            sign_redirect = False if binding == BINDING_HTTP_POST and sign else sign

            reqid, request = self.create_authn_request(
                destination,
                vorg,
                scoping,
                response_binding,
                nameid_format,
                consent=consent,
                extensions=extensions,
                sign=sign_post,
                sign_alg=sigalg,
                digest_alg=digest_alg,
                **kwargs,
            )

            _req_str = str(request)
            logger.info("AuthNReq: %s", _req_str)

            http_info = self.apply_binding(
                binding,
                _req_str,
                destination,
                relay_state,
                sign=sign_redirect,
                sigalg=sigalg,
            )

            return reqid, binding, http_info
        else:
            raise SignOnError("No supported bindings available for authentication")

    def global_logout(
        self,
        name_id,
        reason="",
        expire=None,
        sign=None,
        sign_alg=None,
        digest_alg=None,
    ):
        """ More or less a layer of indirection :-/
        Bootstrapping the whole thing by finding all the IdPs that should
        be notified.

        :param name_id: The identifier of the subject that wants to be
            logged out.
        :param reason: Why the subject wants to log out
        :param expire: The latest the log out should happen.
            If this time has passed don't bother.
        :param sign: Whether the request should be signed or not.
            This also depends on what binding is used.
        :return: Depends on which binding is used:
            If the HTTP redirect binding then a HTTP redirect,
            if SOAP binding has been used the just the result of that
            conversation.
        """

        if isinstance(name_id, six.string_types):
            name_id = decode(name_id)

        logger.info("logout request for: %s", name_id)

        # find out which IdPs/AAs I should notify
        entity_ids = self.users.issuers_of_info(name_id)
        return self.do_logout(
            name_id,
            entity_ids,
            reason,
            expire,
            sign,
            sign_alg=sign_alg,
            digest_alg=digest_alg,
        )

    def do_logout(
        self,
        name_id,
        entity_ids,
        reason,
        expire,
        sign=None,
        expected_binding=None,
        sign_alg=None,
        digest_alg=None,
        **kwargs,
    ):
        """

        :param name_id: Identifier of the Subject (a NameID instance)
        :param entity_ids: List of entity ids for the IdPs that have provided
            information concerning the subject
        :param reason: The reason for doing the logout
        :param expire: Try to logout before this time.
        :param sign: Whether to sign the request or not
        :param expected_binding: Specify the expected binding then not try it
            all
        :param kwargs: Extra key word arguments.
        :return:
        """
        # check time
        if not not_on_or_after(expire):  # I've run out of time
            # Do the local logout anyway
            self.local_logout(name_id)
            return 0, "504 Gateway Timeout", [], []

        not_done = entity_ids[:]
        responses = {}

        for entity_id in entity_ids:
            logger.debug("Logout from '%s'", entity_id)
            # for all where I can use the SOAP binding, do those first
            for binding in [BINDING_SOAP, BINDING_HTTP_POST, BINDING_HTTP_REDIRECT]:
                if expected_binding and binding != expected_binding:
                    continue

                try:
                    srvs = self.metadata.single_logout_service(
                        entity_id, binding, "idpsso"
                    )
                except:
                    srvs = None

                if not srvs:
                    logger.debug("No SLO '%s' service", binding)
                    continue

                destination = next(locations(srvs), None)
                logger.info("destination to provider: %s", destination)

                try:
                    session_info = self.users.get_info_from(
                        name_id, entity_id, False
                    )
                    session_indexes = [session_info['session_index']]
                except KeyError:
                    session_indexes = None

                sign_post = False if binding == BINDING_HTTP_REDIRECT else sign
                sign_redirect = False if binding == BINDING_HTTP_POST and sign else sign

                req_id, request = self.create_logout_request(
                    destination,
                    entity_id,
                    name_id=name_id,
                    reason=reason,
                    expire=expire,
                    session_indexes=session_indexes,
                    sign=sign_post,
                    sign_alg=sign_alg,
                    digest_alg=digest_alg,
                )

                relay_state = self._relay_state(req_id)
                http_info = self.apply_binding(
                    binding,
                    str(request),
                    destination,
                    relay_state,
                    sign=sign_redirect,
                    sigalg=sign_alg,
                )

                if binding == BINDING_SOAP:
                    response = self.send(**http_info)
                    if response and response.status_code == 200:
                        not_done.remove(entity_id)
                        response = response.text
                        logger.info("Response: %s", response)
                        res = self.parse_logout_request_response(response, binding)
                        responses[entity_id] = res
                    else:
                        logger.info("NOT OK response from %s", destination)
                else:
                    self.state[req_id] = {
                        "entity_id": entity_id,
                        "operation": "SLO",
                        "entity_ids": entity_ids,
                        "name_id": code(name_id),
                        "reason": reason,
                        "not_on_or_after": expire,
                        "sign": sign,
                    }
                    responses[entity_id] = (binding, http_info)
                    not_done.remove(entity_id)

                # only try one binding
                break

        if not_done:
            # upstream should try later
            raise LogoutError("%s" % (entity_ids,))

        return responses

    def local_logout(self, name_id):
        """ Remove the user from the cache, equals local logout

        :param name_id: The identifier of the subject
        """
        self.users.remove_person(name_id)
        return True

    def is_logged_in(self, name_id):
        """ Check if user is in the cache

        :param name_id: The identifier of the subject
        """
        identity = self.users.get_identity(name_id)[0]
        return bool(identity)

    def handle_logout_response(self, response, sign_alg=None, digest_alg=None):
        """ handles a Logout response

        :param response: A response.Response instance
        :return: 4-tuple of (session_id of the last sent logout request,
            response message, response headers and message)
        """

        logger.info("state: %s", self.state)
        status = self.state[response.in_response_to]
        logger.info("status: %s", status)
        issuer = response.issuer()
        logger.info("issuer: %s", issuer)
        del self.state[response.in_response_to]
        if status["entity_ids"] == [issuer]:  # done
            self.local_logout(decode(status["name_id"]))
            return 0, "200 Ok", [("Content-type", "text/html")], []
        else:
            status["entity_ids"].remove(issuer)
            if "sign_alg" in status:
                sign_alg = status["sign_alg"]
            return self.do_logout(
                decode(status["name_id"]),
                status["entity_ids"],
                status["reason"],
                status["not_on_or_after"],
                status["sign"],
                sign_alg=sign_alg,
                digest_alg=digest_alg,
            )

    def _use_soap(self, destination, query_type, **kwargs):
        _create_func = getattr(self, "create_%s" % query_type)
        _response_func = getattr(self, "parse_%s_response" % query_type)
        try:
            response_args = kwargs["response_args"]
            del kwargs["response_args"]
        except KeyError:
            response_args = None

        qid, query = _create_func(destination, **kwargs)

        response = self.send_using_soap(query, destination)

        if response.status_code == 200:
            if not response_args:
                response_args = {"binding": BINDING_SOAP}
            else:
                response_args["binding"] = BINDING_SOAP

            logger.info("Verifying response")
            if response_args:
                response = _response_func(response.content, **response_args)
            else:
                response = _response_func(response.content)
        else:
            raise HTTPError("%d:%s" % (response.status_code, response.error))

        if response:
            # not_done.remove(entity_id)
            logger.info("OK response from %s", destination)
            return response
        else:
            logger.info("NOT OK response from %s", destination)

        return None

    # noinspection PyUnusedLocal
    def do_authz_decision_query(self, entity_id, action,
                                subject_id, nameid_format,
                                evidence=None, resource=None,
                                sp_name_qualifier=None,
                                name_qualifier=None,
                                consent=None, extensions=None, sign=False):

        subject = saml.Subject(
            name_id=saml.NameID(text=subject_id, format=nameid_format,
                                sp_name_qualifier=sp_name_qualifier,
                                name_qualifier=name_qualifier))

        srvs = self.metadata.authz_service(entity_id, BINDING_SOAP)
        for dest in locations(srvs):
            resp = self._use_soap(dest, "authz_decision_query",
                                  action=action, evidence=evidence,
                                  resource=resource, subject=subject)
            if resp:
                return resp

        return None

    def do_assertion_id_request(self, assertion_ids, entity_id,
                                consent=None, extensions=None, sign=False):

        srvs = self.metadata.assertion_id_request_service(entity_id,
                                                          BINDING_SOAP)
        if not srvs:
            raise NoServiceDefined("%s: %s" % (entity_id,
                                               "assertion_id_request_service"))

        if isinstance(assertion_ids, six.string_types):
            assertion_ids = [assertion_ids]

        _id_refs = [AssertionIDRef(_id) for _id in assertion_ids]

        for destination in locations(srvs):
            res = self._use_soap(destination, "assertion_id_request",
                                 assertion_id_refs=_id_refs, consent=consent,
                                 extensions=extensions, sign=sign)
            if res:
                return res

        return None

    def do_authn_query(self, entity_id,
                       consent=None, extensions=None, sign=False):

        srvs = self.metadata.authn_request_service(entity_id, BINDING_SOAP)

        for destination in locations(srvs):
            resp = self._use_soap(destination, "authn_query", consent=consent,
                                  extensions=extensions, sign=sign)
            if resp:
                return resp

        return None

    def do_attribute_query(
        self,
        entityid,
        subject_id,
        attribute=None,
        sp_name_qualifier=None,
        name_qualifier=None,
        nameid_format=None,
        real_id=None,
        consent=None,
        extensions=None,
        sign=False,
        binding=BINDING_SOAP,
        nsprefix=None,
        sign_alg=None,
        digest_alg=None,
    ):
        """ Does a attribute request to an attribute authority, this is
        by default done over SOAP.

        :param entityid: To whom the query should be sent
        :param subject_id: The identifier of the subject
        :param attribute: A dictionary of attributes and values that is
            asked for
        :param sp_name_qualifier: The unique identifier of the
            service provider or affiliation of providers for whom the
            identifier was generated.
        :param name_qualifier: The unique identifier of the identity
            provider that generated the identifier.
        :param nameid_format: The format of the name ID
        :param real_id: The identifier which is the key to this entity in the
            identity database
        :param binding: Which binding to use
        :param nsprefix: Namespace prefixes preferred before those automatically
            produced.
        :return: The attributes returned if BINDING_SOAP was used.
            HTTP args if BINDING_HTT_POST was used.
        """

        if real_id:
            response_args = {"real_id": real_id}
        else:
            response_args = {}

        if not binding:
            binding, destination = self.pick_binding(
                "attribute_service", None, "attribute_authority", entity_id=entityid
            )
        else:
            srvs = self.metadata.attribute_service(entityid, binding)
            if srvs is []:
                raise SAMLError("No attribute service support at entity")

            destination = next(locations(srvs), None)

        if binding == BINDING_SOAP:
            return self._use_soap(
                destination,
                "attribute_query",
                consent=consent,
                extensions=extensions,
                sign=sign,
                sign_alg=sign_alg,
                digest_alg=digest_alg,
                subject_id=subject_id,
                attribute=attribute,
                sp_name_qualifier=sp_name_qualifier,
                name_qualifier=name_qualifier,
                format=nameid_format,
                response_args=response_args,
            )
        elif binding == BINDING_HTTP_POST:
            mid = sid()
            query = self.create_attribute_query(
                destination,
                name_id=subject_id,
                attribute=attribute,
                message_id=mid,
                consent=consent,
                extensions=extensions,
                sign=sign,
                sign_alg=sign_alg,
                digest_alg=digest_alg,
                nsprefix=nsprefix,
            )
            self.state[query.id] = {
                "entity_id": entityid,
                "operation": "AttributeQuery",
                "subject_id": subject_id,
                "sign": sign,
            }
            relay_state = self._relay_state(query.id)
            return self.apply_binding(
                binding,
                str(query),
                destination,
                relay_state,
                sign=False,
                sigalg=sign_alg,
            )
        else:
            raise SAMLError("Unsupported binding")

    def handle_logout_request(
        self,
        request,
        name_id,
        binding,
        sign=None,
        sign_alg=None,
        digest_alg=None,
        relay_state="",
    ):
        """
        Deal with a LogoutRequest

        :param request: The request as text string
        :param name_id: The id of the current user
        :param binding: Which binding the message came in over
        :param sign: Whether the response will be signed or not
        :return: Keyword arguments which can be used to send the response
            what's returned follow different patterns for different bindings.
            If the binding is BINDIND_SOAP, what is returned looks like this::

                {
                    "data": <the SOAP enveloped response>
                    "url": "",
                    'headers': [('content-type', 'application/soap+xml')]
                    'method': "POST
                }
        """
        logger.info("logout request: %s", request)

        _req = self._parse_request(request, LogoutRequest,
                                   "single_logout_service", binding)

        if _req.message.name_id == name_id:
            try:
                if self.local_logout(name_id):
                    status = success_status_factory()
                else:
                    status = status_message_factory("Server error",
                                                    STATUS_REQUEST_DENIED)
            except KeyError:
                status = status_message_factory("Server error",
                                                STATUS_REQUEST_DENIED)
        else:
            status = status_message_factory("Wrong user",
                                            STATUS_UNKNOWN_PRINCIPAL)

        if binding == BINDING_SOAP:
            response_bindings = [BINDING_SOAP]
        elif binding in [BINDING_HTTP_POST, BINDING_HTTP_REDIRECT]:
            response_bindings = [BINDING_HTTP_POST, BINDING_HTTP_REDIRECT]
        else:
            response_bindings = self.config.preferred_binding["single_logout_service"]

        if sign is None:
            sign = self.logout_responses_signed

        response = self.create_logout_response(
            _req.message,
            bindings=response_bindings,
            status=status,
            sign=sign,
            sign_alg=sign_alg,
            digest_alg=digest_alg,
        )
        rinfo = self.response_args(_req.message, response_bindings)

        return self.apply_binding(
            rinfo["binding"],
            response,
            rinfo["destination"],
            relay_state,
            response=True,
            sign=sign,
            sigalg=sign_alg,
        )



#!/usr/bin/env python

#
# Generated Fri May 27 17:30:44 2011 by parse_xsd.py version 0.4.
#

import saml2
from saml2 import SamlBase

#import soapenv as S

NAMESPACE = 'urn:liberty:paos:2003-08'

class RequestType_(SamlBase):
    """The urn:liberty:paos:2003-08:RequestType element """

    c_tag = 'RequestType'
    c_namespace = NAMESPACE
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()
    c_attributes['responseConsumerURL'] = ('response_consumer_url', 'anyURI', True)
    c_attributes['service'] = ('service', 'anyURI', True)
    c_attributes['messageID'] = ('message_id', 'None', False)
    c_attributes['{http://schemas.xmlsoap.org/soap/envelope/}mustUnderstand'] = ('must_understand', 'None', True)
    c_attributes['{http://schemas.xmlsoap.org/soap/envelope/}actor'] = ('actor', 'None', True)

    def __init__(self,
            response_consumer_url=None,
            service=None,
            message_id=None,
            must_understand=None,
            actor=None,
            text=None,
            extension_elements=None,
            extension_attributes=None,
        ):
        SamlBase.__init__(self,
                text=text,
                extension_elements=extension_elements,
                extension_attributes=extension_attributes,
                )
        self.response_consumer_url=response_consumer_url
        self.service=service
        self.message_id=message_id
        self.must_understand=must_understand
        self.actor=actor

def request_type__from_string(xml_string):
    return saml2.create_class_from_xml_string(RequestType_, xml_string)


class ResponseType_(SamlBase):
    """The urn:liberty:paos:2003-08:ResponseType element """

    c_tag = 'ResponseType'
    c_namespace = NAMESPACE
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()
    c_attributes['refToMessageID'] = ('ref_to_message_id', 'None', False)
    c_attributes['{http://schemas.xmlsoap.org/soap/envelope/}mustUnderstand'] = ('must_understand', 'None', True)
    c_attributes['{http://schemas.xmlsoap.org/soap/envelope/}actor'] = ('actor', 'None', True)

    def __init__(self,
            ref_to_message_id=None,
            must_understand=None,
            actor=None,
            text=None,
            extension_elements=None,
            extension_attributes=None,
        ):
        SamlBase.__init__(self,
                text=text,
                extension_elements=extension_elements,
                extension_attributes=extension_attributes,
                )
        self.ref_to_message_id=ref_to_message_id
        self.must_understand=must_understand
        self.actor=actor

def response_type__from_string(xml_string):
    return saml2.create_class_from_xml_string(ResponseType_, xml_string)


class Request(RequestType_):
    """The urn:liberty:paos:2003-08:Request element """

    c_tag = 'Request'
    c_namespace = NAMESPACE
    c_children = RequestType_.c_children.copy()
    c_attributes = RequestType_.c_attributes.copy()
    c_child_order = RequestType_.c_child_order[:]
    c_cardinality = RequestType_.c_cardinality.copy()

def request_from_string(xml_string):
    return saml2.create_class_from_xml_string(Request, xml_string)


class Response(ResponseType_):
    """The urn:liberty:paos:2003-08:Response element """

    c_tag = 'Response'
    c_namespace = NAMESPACE
    c_children = ResponseType_.c_children.copy()
    c_attributes = ResponseType_.c_attributes.copy()
    c_child_order = ResponseType_.c_child_order[:]
    c_cardinality = ResponseType_.c_cardinality.copy()

def response_from_string(xml_string):
    return saml2.create_class_from_xml_string(Response, xml_string)


ELEMENT_FROM_STRING = {
    Request.c_tag: request_from_string,
    RequestType_.c_tag: request_type__from_string,
    Response.c_tag: response_from_string,
    ResponseType_.c_tag: response_type__from_string,
}

ELEMENT_BY_TAG = {
    'Request': Request,
    'RequestType': RequestType_,
    'Response': Response,
    'ResponseType': ResponseType_,
}


def factory(tag, **kwargs):
    return ELEMENT_BY_TAG[tag](**kwargs)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

"""Contains classes and functions that a SAML2.0 Identity provider (IdP)
or attribute authority (AA) may use to conclude its tasks.
"""
import logging
import os

import importlib
import dbm
import shelve
import six
import threading

import saml2.cryptography.symmetric
from saml2 import saml
from saml2 import element_to_extension_element
from saml2 import class_name
from saml2 import BINDING_HTTP_REDIRECT
from saml2.argtree import add_path, is_set

from saml2.entity import Entity
from saml2.eptid import Eptid
from saml2.eptid import EptidShelve
from saml2.samlp import NameIDMappingResponse
from saml2.sdb import SessionStorage
from saml2.schema import soapenv

from saml2.request import AuthnRequest
from saml2.request import AssertionIDRequest
from saml2.request import AttributeQuery
from saml2.request import NameIDMappingRequest
from saml2.request import AuthzDecisionQuery
from saml2.request import AuthnQuery

from saml2.s_utils import MissingValue
from saml2.s_utils import rndstr
from saml2.s_utils import Unknown

from saml2.sigver import pre_signature_part
from saml2.sigver import signed_instance_factory
from saml2.sigver import CertificateError

from saml2.assertion import Assertion
from saml2.assertion import Policy
from saml2.assertion import restriction_from_attribute_spec
from saml2.assertion import filter_attribute_value_assertions

from saml2.ident import IdentDB, decode
from saml2.profile import ecp

logger = logging.getLogger(__name__)

AUTHN_DICT_MAP = {
    "decl": "authn_decl",
    "authn_auth": "authn_auth",
    "class_ref": "authn_class",
    "authn_instant": "authn_instant",
    "subject_locality": "subject_locality"
}


def _shelve_compat(name, *args, **kwargs):
    try:
        return shelve.open(name, *args, **kwargs)
    except dbm.error[0]:
        # Python 3 whichdb needs to try .db to determine type
        if name.endswith('.db'):
            name = name.rsplit('.db', 1)[0]
            return shelve.open(name, *args, **kwargs)
        else:
            raise


class Server(Entity):
    """ A class that does things that IdPs or AAs do """

    def __init__(
        self,
        config_file="",
        config=None,
        cache=None,
        stype="idp",
        symkey="",
        msg_cb=None,
    ):
        Entity.__init__(self, stype, config, config_file, msg_cb=msg_cb)
        self.eptid = None
        self.init_config(stype)
        self.cache = cache
        self.ticket = {}
        self.session_db = self.choose_session_storage()
        if symkey:
            self.symkey = symkey.encode()
        else:
            self.symkey = saml2.cryptography.symmetric.Default.generate_key()
        self.seed = rndstr()
        self.lock = threading.Lock()

    def getvalid_certificate_str(self):
        if self.sec.cert_handler is not None:
            return self.sec.cert_handler._last_validated_cert
        return None

    def support_AssertionIDRequest(self):
        return True

    def support_AuthnQuery(self):
        return True

    def choose_session_storage(self):
        _spec = self.config.getattr("session_storage", "idp")
        if not _spec:
            return SessionStorage()
        elif isinstance(_spec, six.string_types):
            if _spec.lower() == "memory":
                return SessionStorage()
        else:  # Should be tuple
            typ, data = _spec
            if typ.lower() == "mongodb":
                from saml2.mongo_store import SessionStorageMDB
                return SessionStorageMDB(database=data, collection="session")

        raise NotImplementedError("No such storage type implemented")

    def init_config(self, stype="idp"):
        """ Remaining init of the server configuration

        :param stype: The type of Server ("idp"/"aa")
        """
        if stype == "aa":
            return

        # subject information is stored in a database
        # default database is in memory which is OK in some setups
        dbspec = self.config.getattr("subject_data", "idp")
        idb = None
        typ = ""
        if not dbspec:
            idb = {}
        elif isinstance(dbspec, six.string_types):
            idb = _shelve_compat(dbspec, writeback=True, protocol=2)
        else:  # database spec is a a 2-tuple (type, address)
            # print(>> sys.stderr, "DBSPEC: %s" % (dbspec,))
            (typ, addr) = dbspec
            if typ == "shelve":
                idb = _shelve_compat(addr, writeback=True, protocol=2)
            elif typ == "memcached":
                import memcache
                idb = memcache.Client(addr)
            elif typ == "dict":  # in-memory dictionary
                idb = {}
            elif typ == "mongodb":
                from saml2.mongo_store import IdentMDB
                self.ident = IdentMDB(database=addr, collection="ident")
            elif typ == "identdb":
                mod, clas = addr.rsplit('.', 1)
                mod = importlib.import_module(mod)
                self.ident = getattr(mod, clas)()

        if typ == "mongodb" or typ == "identdb":
            pass
        elif idb is not None:
            self.ident = IdentDB(idb)
        elif dbspec:
            raise Exception("Couldn't open identity database: %s" %
                            (dbspec,))

        try:
            _domain = self.config.getattr("domain", "idp")
            if _domain:
                self.ident.domain = _domain

            self.ident.name_qualifier = self.config.entityid

            dbspec = self.config.getattr("edu_person_targeted_id", "idp")
            if not dbspec:
                pass
            else:
                typ = dbspec[0]
                addr = dbspec[1]
                secret = dbspec[2]
                if typ == "shelve":
                    self.eptid = EptidShelve(secret, addr)
                elif typ == "mongodb":
                    from saml2.mongo_store import EptidMDB
                    self.eptid = EptidMDB(secret, database=addr,
                                          collection="eptid")
                else:
                    self.eptid = Eptid(secret)
        except Exception:
            self.ident.close()
            raise

    def wants(self, sp_entity_id, index=None):
        """ Returns what attributes the SP requires and which are optional
        if any such demands are registered in the Metadata.

        :param sp_entity_id: The entity id of the SP
        :param index: which of the attribute consumer services its all about
            if index == None then all attribute consumer services are clumped
            together.
        :return: 2-tuple, list of required and list of optional attributes
        """
        return self.metadata.attribute_requirement(sp_entity_id, index)

    def verify_assertion_consumer_service(self, request):
        _acs = request.assertion_consumer_service_url
        _aci = request.assertion_consumer_service_index
        _binding = request.protocol_binding
        _eid = request.issuer.text
        if _acs:
            # look up acs in for that binding in the metadata given the issuer
            # Assuming the format is entity
            for acs in self.metadata.assertion_consumer_service(_eid, _binding):
                if _acs == acs.text:
                    return True
        elif _aci:
            for acs in self.metadata.assertion_consumer_service(_eid, _binding):
                if _aci == acs.index:
                    return True

        return False

    # -------------------------------------------------------------------------

    def parse_authn_request(self, enc_request, binding=BINDING_HTTP_REDIRECT):
        """Parse a Authentication Request

        :param enc_request: The request in its transport format
        :param binding: Which binding that was used to transport the message
            to this entity.
        :return: A request instance
        """

        return self._parse_request(enc_request, AuthnRequest,
                                   "single_sign_on_service", binding)

    def parse_attribute_query(self, xml_string, binding):
        """ Parse an attribute query

        :param xml_string: The Attribute Query as an XML string
        :param binding: Which binding that was used for the request
        :return: A query instance
        """

        return self._parse_request(xml_string, AttributeQuery,
                                   "attribute_service", binding)

    def parse_authz_decision_query(self, xml_string, binding):
        """ Parse an authorization decision query

        :param xml_string: The Authz decision Query as an XML string
        :param binding: Which binding that was used when receiving this query
        :return: Query instance
        """

        return self._parse_request(xml_string, AuthzDecisionQuery,
                                   "authz_service", binding)

    def parse_assertion_id_request(self, xml_string, binding):
        """ Parse an assertion id query

        :param xml_string: The AssertionIDRequest as an XML string
        :param binding: Which binding that was used when receiving this request
        :return: Query instance
        """

        return self._parse_request(xml_string, AssertionIDRequest,
                                   "assertion_id_request_service", binding)

    def parse_authn_query(self, xml_string, binding):
        """ Parse an authn query

        :param xml_string: The AuthnQuery as an XML string
        :param binding: Which binding that was used when receiving this query
        :return: Query instance
        """

        return self._parse_request(xml_string, AuthnQuery,
                                   "authn_query_service", binding)

    def parse_name_id_mapping_request(self, xml_string, binding):
        """ Parse a nameid mapping request

        :param xml_string: The NameIDMappingRequest as an XML string
        :param binding: Which binding that was used when receiving this request
        :return: Query instance
        """

        return self._parse_request(xml_string, NameIDMappingRequest,
                                   "name_id_mapping_service", binding)

    @staticmethod
    def update_farg(in_response_to, consumer_url, farg=None):
        if not farg:
            farg = add_path(
                {},
                ['assertion', 'subject', 'subject_confirmation', 'method',
                 saml.SCM_BEARER])
            add_path(
                farg['assertion']['subject']['subject_confirmation'],
                ['subject_confirmation_data', 'in_response_to', in_response_to])
            add_path(
                farg['assertion']['subject']['subject_confirmation'],
                ['subject_confirmation_data', 'recipient', consumer_url])
        else:
            if not is_set(farg,
                          ['assertion', 'subject', 'subject_confirmation',
                           'method']):
                add_path(farg,
                         ['assertion', 'subject', 'subject_confirmation',
                          'method', saml.SCM_BEARER])
            if not is_set(farg,
                          ['assertion', 'subject', 'subject_confirmation',
                           'subject_confirmation_data', 'in_response_to']):
                add_path(farg,
                         ['assertion', 'subject', 'subject_confirmation',
                          'subject_confirmation_data', 'in_response_to',
                          in_response_to])
            if not is_set(farg, ['assertion', 'subject', 'subject_confirmation',
                                 'subject_confirmation_data', 'recipient']):
                add_path(farg,
                         ['assertion', 'subject', 'subject_confirmation',
                          'subject_confirmation_data', 'recipient',
                          consumer_url])
        return farg

    def setup_assertion(
        self,
        authn,
        sp_entity_id,
        in_response_to,
        consumer_url,
        name_id,
        policy,
        _issuer,
        authn_statement,
        identity,
        best_effort,
        sign_response,
        farg=None,
        session_not_on_or_after=None,
        sign_alg=None,
        digest_alg=None,
        **kwargs,
    ):
        """
        Construct and return the Assertion

        :param authn: Authentication information
        :param sp_entity_id:
        :param in_response_to: The ID of the request this is an answer to
        :param consumer_url: The recipient of the assertion
        :param name_id: The NameID of the subject
        :param policy: Assertion policies
        :param _issuer: Issuer of the statement
        :param authn_statement: An AuthnStatement instance
        :param identity: Identity information about the Subject
        :param best_effort: Even if not the SPs demands can be met send a
            response.
        :param sign_response: Sign the response, only applicable if
            ErrorResponse
        :param kwargs: Extra keyword arguments
        :return: An Assertion instance
        """

        ast = Assertion(identity)
        ast.acs = self.config.getattr("attribute_converters")
        if policy is None:
            policy = Policy(mds=self.metadata)
        try:
            ast.apply_policy(sp_entity_id, policy)
        except MissingValue as exc:
            if not best_effort:
                response = self.create_error_response(
                    in_response_to,
                    destination=consumer_url,
                    info=exc,
                    sign=sign_response,
                    sign_alg=sign_alg,
                    digest_alg=digest_alg,
                )
                return str(response).split("\n")

        farg = self.update_farg(in_response_to, consumer_url, farg)

        if authn:  # expected to be a dictionary
            # Would like to use dict comprehension but ...
            authn_args = dict(
                [(AUTHN_DICT_MAP[k], v) for k, v in authn.items() if
                 k in AUTHN_DICT_MAP])
            authn_args.update(kwargs)

            assertion = ast.construct(
                sp_entity_id, self.config.attribute_converters, policy,
                issuer=_issuer, farg=farg['assertion'], name_id=name_id,
                session_not_on_or_after=session_not_on_or_after,
                **authn_args)

        elif authn_statement:  # Got a complete AuthnStatement
            assertion = ast.construct(
                sp_entity_id, self.config.attribute_converters, policy,
                issuer=_issuer, authn_statem=authn_statement,
                farg=farg['assertion'], name_id=name_id,
                **kwargs)
        else:
            assertion = ast.construct(
                sp_entity_id, self.config.attribute_converters, policy,
                issuer=_issuer, farg=farg['assertion'], name_id=name_id,
                session_not_on_or_after=session_not_on_or_after,
                **kwargs)
        return assertion

    # XXX DONE calls pre_signature_part
    # XXX calls _response
    def _authn_response(
        self,
        in_response_to,
        consumer_url,
        sp_entity_id,
        identity=None,
        name_id=None,
        status=None,
        authn=None,
        issuer=None,
        policy=None,
        sign_assertion=None,
        sign_response=None,
        best_effort=False,
        encrypt_assertion=False,
        encrypt_cert_advice=None,
        encrypt_cert_assertion=None,
        authn_statement=None,
        encrypt_assertion_self_contained=False,
        encrypted_advice_attributes=False,
        pefim=False,
        sign_alg=None,
        digest_alg=None,
        farg=None,
        session_not_on_or_after=None,
    ):
        """ Create a response. A layer of indirection.

        :param in_response_to: The session identifier of the request
        :param consumer_url: The URL which should receive the response
        :param sp_entity_id: The entity identifier of the SP
        :param identity: A dictionary with attributes and values that are
            expected to be the bases for the assertion in the response.
        :param name_id: The identifier of the subject
        :param status: The status of the response
        :param authn: A dictionary containing information about the
            authn context.
        :param issuer: The issuer of the response
        :param policy:
        :param sign_assertion: Whether the assertion should be signed or not
        :param sign_response: Whether the response should be signed or not
        :param best_effort: Even if not the SPs demands can be met send a
            response.
        :param encrypt_assertion: True if assertions should be encrypted.
        :param encrypt_assertion_self_contained: True if all encrypted
        assertions should have alla namespaces
        selfcontained.
        :param encrypted_advice_attributes: True if assertions in the advice
        element should be encrypted.
        :param encrypt_cert_advice: Certificate to be used for encryption of
        assertions in the advice element.
        :param encrypt_cert_assertion: Certificate to be used for encryption
        of assertions.
        :param authn_statement: Authentication statement.
        :param pefim: True if a response according to the PEFIM profile
        should be created.
        :param farg: Argument to pass on to the assertion constructor
        :return: A response instance
        """

        if farg is None:
            assertion_args = {}

        # if identity:
        _issuer = self._issuer(issuer)

        # if encrypt_assertion and show_nameid:
        #    tmp_name_id = name_id
        #    name_id = None
        #    name_id = None
        #    tmp_authn = authn
        #    authn = None
        #    tmp_authn_statement = authn_statement
        #    authn_statement = None

        if pefim:
            encrypted_advice_attributes = True
            encrypt_assertion_self_contained = True
            assertion_attributes = self.setup_assertion(
                None, sp_entity_id, None, None, None, policy, None, None,
                identity, best_effort, sign_response, farg=farg)
            assertion = self.setup_assertion(
                authn, sp_entity_id, in_response_to, consumer_url, name_id,
                policy, _issuer, authn_statement, [], True, sign_response,
                farg=farg, session_not_on_or_after=session_not_on_or_after)
            assertion.advice = saml.Advice()

            # assertion.advice.assertion_id_ref.append(saml.AssertionIDRef())
            # assertion.advice.assertion_uri_ref.append(saml.AssertionURIRef())
            assertion.advice.assertion.append(assertion_attributes)
        else:
            assertion = self.setup_assertion(
                authn, sp_entity_id, in_response_to, consumer_url, name_id,
                policy, _issuer, authn_statement, identity, True,
                sign_response, farg=farg,
                session_not_on_or_after=session_not_on_or_after)

        to_sign = []
        if not encrypt_assertion:
            if sign_assertion:
                # XXX self.signing_algorithm self.digest_algorithm defined by entity
                # XXX this should be handled through entity.py
                # XXX sig/digest-allowed should be configurable
                sign_alg = sign_alg or self.signing_algorithm
                digest_alg = digest_alg or self.digest_algorithm

                assertion.signature = pre_signature_part(
                    assertion.id,
                    self.sec.my_cert,
                    2,
                    sign_alg=sign_alg,
                    digest_alg=digest_alg,
                )
                to_sign.append((class_name(assertion), assertion.id))

        if (self.support_AssertionIDRequest() or self.support_AuthnQuery()):
            self.session_db.store_assertion(assertion, to_sign)

        return self._response(
            in_response_to,
            consumer_url,
            status,
            issuer,
            sign_response,
            to_sign,
            sp_entity_id=sp_entity_id,
            encrypt_assertion=encrypt_assertion,
            encrypt_cert_advice=encrypt_cert_advice,
            encrypt_cert_assertion=encrypt_cert_assertion,
            encrypt_assertion_self_contained=encrypt_assertion_self_contained,
            encrypted_advice_attributes=encrypted_advice_attributes,
            sign_assertion=sign_assertion,
            pefim=pefim,
            sign_alg=sign_alg,
            digest_alg=digest_alg,
            assertion=assertion,
        )

    # ------------------------------------------------------------------------

    # XXX DONE idp create > _response
    def create_attribute_response(
        self,
        identity,
        in_response_to,
        destination,
        sp_entity_id,
        userid="",
        name_id=None,
        status=None,
        issuer=None,
        sign_assertion=None,
        sign_response=None,
        attributes=None,
        sign_alg=None,
        digest_alg=None,
        farg=None,
        **kwargs,
    ):
        """ Create an attribute assertion response.

        :param identity: A dictionary with attributes and values that are
            expected to be the bases for the assertion in the response.
        :param in_response_to: The session identifier of the request
        :param destination: The URL which should receive the response
        :param sp_entity_id: The entity identifier of the SP
        :param userid: A identifier of the user
        :param name_id: The identifier of the subject
        :param status: The status of the response
        :param issuer: The issuer of the response
        :param sign_assertion: Whether the assertion should be signed or not
        :param sign_response: Whether the whole response should be signed
        :param attributes:
        :param kwargs: To catch extra keyword arguments
        :return: A response instance
        """

        policy = self.config.getattr("policy", "aa")

        if not name_id and userid:
            try:
                name_id = self.ident.construct_nameid(userid, policy, sp_entity_id)
                logger.warning("Unspecified NameID format")
            except Exception:
                pass

        to_sign = []

        if identity:
            farg = self.update_farg(in_response_to, sp_entity_id, farg=farg)

            _issuer = self._issuer(issuer)
            ast = Assertion(identity)
            if policy:
                ast.apply_policy(sp_entity_id, policy)
            else:
                policy = Policy(mds=self.metadata)

            if attributes:
                restr = restriction_from_attribute_spec(attributes)
                ast = filter_attribute_value_assertions(ast)

            assertion = ast.construct(
                sp_entity_id, self.config.attribute_converters, policy,
                issuer=_issuer, name_id=name_id,
                farg=farg['assertion'])

        return self._response(
            in_response_to,
            destination,
            status,
            issuer,
            sign_response,
            to_sign,
            sign_assertion=sign_assertion,
            sign_alg=sign_alg,
            digest_alg=digest_alg,
            assertion=assertion,
            sp_entity_id=sp_entity_id,
            **kwargs,
        )

    def gather_authn_response_args(
        self, sp_entity_id, name_id_policy, userid, **kwargs
    ):
        kwargs["policy"] = kwargs.get("release_policy")

        # collect args and return them
        args = {}

        # XXX will be passed to _authn_response
        param_defaults = {
            'policy': None,
            'best_effort': False,
            'sign_assertion': False,
            'sign_response': False,
            'encrypt_assertion': False,
            'encrypt_assertion_self_contained': True,
            'encrypted_advice_attributes': False,
            'encrypt_cert_advice': None,
            'encrypt_cert_assertion': None,
            # need to be named sign_alg and digest_alg
        }
        for param, val_default in param_defaults.items():
            val_kw = kwargs.get(param)
            val_config = self.config.getattr(param, "idp")
            args[param] = (
                val_kw
                if val_kw is not None
                else val_config
                if val_config is not None
                else val_default
            )

        for arg, attr, eca, pefim in [
            ('encrypted_advice_attributes', 'verify_encrypt_cert_advice',
             'encrypt_cert_advice', kwargs["pefim"]),
            ('encrypt_assertion', 'verify_encrypt_cert_assertion',
             'encrypt_cert_assertion', False)]:

            if args[arg] or pefim:
                _enc_cert = self.config.getattr(attr, "idp")

                if _enc_cert is not None:
                    if kwargs[eca] is None:
                        raise CertificateError(
                            "No SPCertEncType certificate for encryption "
                            "contained in authentication "
                            "request.")
                    if not _enc_cert(kwargs[eca]):
                        raise CertificateError(
                            "Invalid certificate for encryption!")

        if 'name_id' not in kwargs or not kwargs['name_id']:
            nid_formats = []
            for _sp in self.metadata[sp_entity_id]["spsso_descriptor"]:
                if "name_id_format" in _sp:
                    nid_formats.extend([n["text"] for n in
                                        _sp["name_id_format"]])
            try:
                snq = name_id_policy.sp_name_qualifier
            except AttributeError:
                snq = sp_entity_id

            if not snq:
                snq = sp_entity_id

            kwa = {"sp_name_qualifier": snq}

            try:
                kwa["format"] = name_id_policy.format
            except AttributeError:
                pass

            _nids = self.ident.find_nameid(userid, **kwa)
            # either none or one
            if _nids:
                args['name_id'] = _nids[0]
            else:
                args['name_id'] = self.ident.construct_nameid(
                    userid, args['policy'], sp_entity_id, name_id_policy)
                logger.debug("construct_nameid: %s => %s", userid,
                             args['name_id'])
        else:
            args['name_id'] = kwargs['name_id']

        for param in ['status', 'farg']:
            try:
                args[param] = kwargs[param]
            except KeyError:
                pass

        return args

    # XXX DONE idp create > _authn_response > _response
    def create_authn_response(
        self,
        identity,
        in_response_to,
        destination,
        sp_entity_id,
        name_id_policy=None,
        userid=None,
        name_id=None,
        authn=None,
        issuer=None,
        sign_response=None,
        sign_assertion=None,
        encrypt_cert_advice=None,
        encrypt_cert_assertion=None,
        encrypt_assertion=None,
        encrypt_assertion_self_contained=True,
        encrypted_advice_attributes=False,
        pefim=False,
        sign_alg=None,
        digest_alg=None,
        session_not_on_or_after=None,
        **kwargs,
    ):
        """ Constructs an AuthenticationResponse

        :param identity: Information about an user
        :param in_response_to: The identifier of the authentication request
            this response is an answer to.
        :param destination: Where the response should be sent
        :param sp_entity_id: The entity identifier of the Service Provider
        :param name_id_policy: How the NameID should be constructed
        :param userid: The subject identifier
        :param name_id: The identifier of the subject. A saml.NameID instance.
        :param authn: Dictionary with information about the authentication
            context
        :param issuer: Issuer of the response
        :param sign_assertion: Whether the assertion should be signed or not.
        :param sign_response: Whether the response should be signed or not.
        :param encrypt_assertion: True if assertions should be encrypted.
        :param encrypt_assertion_self_contained: True if all encrypted
        assertions should have alla namespaces
        selfcontained.
        :param encrypted_advice_attributes: True if assertions in the advice
        element should be encrypted.
        :param encrypt_cert_advice: Certificate to be used for encryption of
        assertions in the advice element.
        :param encrypt_cert_assertion: Certificate to be used for encryption
        of assertions.
        :param pefim: True if a response according to the PEFIM profile
        should be created.
        :return: A response instance
        """

        try:
            args = self.gather_authn_response_args(
                sp_entity_id,
                name_id_policy=name_id_policy,
                userid=userid,
                name_id=name_id,
                sign_response=sign_response,
                sign_assertion=sign_assertion,
                encrypt_cert_advice=encrypt_cert_advice,
                encrypt_cert_assertion=encrypt_cert_assertion,
                encrypt_assertion=encrypt_assertion,
                encrypt_assertion_self_contained=encrypt_assertion_self_contained,
                encrypted_advice_attributes=encrypted_advice_attributes,
                pefim=pefim,
                **kwargs,
            )
        except IOError as exc:
            response = self.create_error_response(
                in_response_to,
                destination=destination,
                info=exc,
                sign=sign_response,
                sign_alg=sign_alg,
                digest_alg=digest_alg,
            )
            return str(response).split("\n")

        try:
            _authn = authn
            return self._authn_response(
                in_response_to,
                destination,
                sp_entity_id,
                identity,
                authn=_authn,
                issuer=issuer,
                pefim=pefim,
                sign_alg=sign_alg,
                digest_alg=digest_alg,
                session_not_on_or_after=session_not_on_or_after,
                **args,
            )
        except MissingValue as exc:
            return self.create_error_response(
                in_response_to,
                destination=destination,
                info=exc,
                sign=sign_response,
                sign_alg=sign_alg,
                digest_alg=digest_alg,
            )

    # XXX DONE idp create > create_authn_response > _authn_response > _response
    def create_authn_request_response(
        self,
        identity,
        in_response_to,
        destination,
        sp_entity_id,
        name_id_policy=None,
        userid=None,
        name_id=None,
        authn=None,
        authn_decl=None,
        issuer=None,
        sign_response=None,
        sign_assertion=None,
        session_not_on_or_after=None,
        sign_alg=None,
        digest_alg=None,
        **kwargs,
    ):
        return self.create_authn_response(
            identity,
            in_response_to,
            destination,
            sp_entity_id,
            name_id_policy,
            userid,
            name_id,
            authn,
            issuer,
            sign_response,
            sign_assertion,
            authn_decl=authn_decl,
            session_not_on_or_after=session_not_on_or_after,
            sign_alg=sign_alg,
            digest_alg=digest_alg,
        )

    # XXX DONE calls pre_signature_part
    # XXX DONE idp create > [...]
    def create_assertion_id_request_response(
        self, assertion_id, sign=None, sign_alg=None, digest_alg=None, **kwargs
    ):
        try:
            (assertion, to_sign) = self.session_db.get_assertion(assertion_id)
        except KeyError:
            raise Unknown

        if to_sign:
            if assertion.signature is None:
                # XXX self.signing_algorithm self.digest_algorithm defined by entity
                # XXX this should be handled through entity.py
                # XXX sig/digest-allowed should be configurable
                sign_alg = sign_alg or self.signing_algorithm
                digest_alg = digest_alg or self.digest_algorithm

                assertion.signature = pre_signature_part(
                    assertion.id,
                    self.sec.my_cert,
                    1,
                    sign_alg=sign_alg,
                    digest_alg=digest_alg,
                )
            return signed_instance_factory(assertion, self.sec, to_sign)
        else:
            return assertion

    # XXX calls self.sign without ensuring sign
    # XXX calls self.sign => should it call _message (which calls self.sign)?
    # XXX idp create > NameIDMappingResponse & sign?
    def create_name_id_mapping_response(
        self,
        name_id=None,
        encrypted_id=None,
        in_response_to=None,
        issuer=None,
        sign_response=None,
        status=None,
        sign_alg=None,
        digest_alg=None,
        **kwargs,
    ):
        """
        protocol for mapping a principal's name identifier into a
        different name identifier for the same principal.
        Done over soap.

        :param name_id:
        :param encrypted_id:
        :param in_response_to:
        :param issuer:
        :param sign_response:
        :param status:
        :return:
        """
        # Done over SOAP

        ms_args = self.message_args()

        _resp = NameIDMappingResponse(
            name_id, encrypted_id, in_response_to=in_response_to, **ms_args
        )

        if sign_response:
            return self.sign(_resp, sign_alg=sign_alg, digest_alg=digest_alg)
        else:
            logger.info("Message: %s", _resp)
            return _resp

    # XXX DONE idp create > _response
    def create_authn_query_response(
        self,
        subject,
        session_index=None,
        requested_context=None,
        in_response_to=None,
        issuer=None,
        sign_response=None,
        status=None,
        sign_alg=None,
        digest_alg=None,
        **kwargs,
    ):
        """
        A successful <Response> will contain one or more assertions containing
        authentication statements.

        :return:
        """

        margs = self.message_args()
        asserts = [
            saml.Assertion(authn_statement=statement, subject=subject, **margs)
            for statement in self.session_db.get_authn_statements(
                subject.name_id, session_index, requested_context
            )
        ]

        if asserts:
            args = {"assertion": asserts}
        else:
            args = {}

        return self._response(
            in_response_to,
            "",
            status,
            issuer,
            sign_response,
            to_sign=[],
            sign_alg=sign_alg,
            digest_alg=digest_alg,
            **args,
        )

    # ---------

    def parse_ecp_authn_request(self):
        pass

    # XXX DONE idp create > create_authn_response > _authn_response > _response
    def create_ecp_authn_request_response(
        self,
        acs_url,
        identity,
        in_response_to,
        destination,
        sp_entity_id,
        name_id_policy=None,
        userid=None,
        name_id=None,
        authn=None,
        issuer=None,
        sign_response=None,
        sign_assertion=None,
        sign_alg=None,
        digest_alg=None,
        **kwargs,
    ):

        # ----------------------------------------
        # <ecp:Response
        # ----------------------------------------

        ecp_response = ecp.Response(assertion_consumer_service_url=acs_url)
        header = soapenv.Header()
        header.extension_elements = [element_to_extension_element(ecp_response)]

        # ----------------------------------------
        # <samlp:Response
        # ----------------------------------------

        response = self.create_authn_response(
            identity,
            in_response_to,
            destination,
            sp_entity_id,
            name_id_policy,
            userid,
            name_id,
            authn,
            issuer,
            sign_response,
            sign_assertion,
            sign_alg=sign_alg,
            digest_alg=digest_alg
        )
        body = soapenv.Body()
        body.extension_elements = [element_to_extension_element(response)]

        soap_envelope = soapenv.Envelope(header=header, body=body)

        return str(soap_envelope)

    def close(self):
        self.ident.close()

    def clean_out_user(self, name_id):
        """
        Remove all authentication statements that belongs to a user identified
        by a NameID instance

        :param name_id: NameID instance
        :return: The local identifier for the user
        """

        lid = self.ident.find_local_id(name_id)
        logger.info("Clean out %s", lid)

        # remove the authentications
        try:
            for _nid in [decode(x) for x in self.ident.db[lid].split(" ")]:
                try:
                    self.session_db.remove_authn_statements(_nid)
                except KeyError:
                    pass
        except KeyError:
            pass

        return lid



"""Setup.py entry point for package.

Configuration is handled by setuptools>30.3.0 through setup.cfg.
https://setuptools.readthedocs.io/en/latest/setuptools.html#metadata
https://setuptools.readthedocs.io/en/latest/setuptools.html#options
"""

import setuptools


setuptools.setup(
    package_dir={'': 'src'},
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

"""
Contains a class that can do SAML ECP Authentication for other python
programs.
"""

from six.moves import http_cookiejar as cookielib
import logging

from saml2 import soap
from saml2 import saml
from saml2 import samlp
from saml2 import SAMLError
from saml2 import BINDING_SOAP
from saml2.client_base import MIME_PAOS
from saml2.config import Config
from saml2.entity import Entity
from saml2.httpbase import set_list2dict, dict2set_list

from saml2.profile import paos
from saml2.profile import ecp

from saml2.mdstore import MetadataStore
from saml2.s_utils import BadRequest

SERVICE = "urn:oasis:names:tc:SAML:2.0:profiles:SSO:ecp"
PAOS_HEADER_INFO = 'ver="%s";"%s"' % (paos.NAMESPACE, SERVICE)

logger = logging.getLogger(__name__)


class Client(Entity):
    """ECP-aware client that works on the client (application) side.

    You can use this class when you want to login user through
    ECP-aware SP and IdP.
    """

    def __init__(self, user, passwd, sp="", idp=None, metadata_file=None,
                 xmlsec_binary=None, verbose=0, ca_certs="",
                 disable_ssl_certificate_validation=True, key_file=None,
                 cert_file=None, config=None):
        """
        :param user: user name
        :param passwd: user password
        :param sp: The SP URL
        :param idp: The IdP PAOS endpoint
        :param metadata_file: Where the metadata file is if used
        :param xmlsec_binary: Where the xmlsec1 binary can be found (*)
        :param verbose: Chatty or not
        :param ca_certs: is the path of a file containing root CA certificates
            for SSL server certificate validation (*)
        :param disable_ssl_certificate_validation: If
            disable_ssl_certificate_validation is true, SSL cert validation
            will not be performed (*)
        :param key_file: Private key filename (*)
        :param cert_file: Certificate filename (*)
        :param config: Config() instance, overrides all the parameters marked
            with an asterisk (*) above
        """
        if not config:
            config = Config()
            config.disable_ssl_certificate_validation = \
                disable_ssl_certificate_validation
            config.key_file = key_file
            config.cert_file = cert_file
            config.ca_certs = ca_certs
            config.xmlsec_binary = xmlsec_binary

        Entity.__init__(self, "sp", config)
        self._idp = idp
        self._sp = sp
        self.user = user
        self.passwd = passwd
        self._verbose = verbose

        if metadata_file:
            self._metadata = MetadataStore([saml, samlp], None, config)
            self._metadata.load("local", metadata_file)
            logger.debug("Loaded metadata from '%s'", metadata_file)
        else:
            self._metadata = None

        self.metadata = self._metadata

        self.cookie_handler = None

        self.done_ecp = False
        self.cookie_jar = cookielib.LWPCookieJar()

    def phase2(
        self,
        authn_request,
        rc_url,
        idp_entity_id,
        headers=None,
        sign=False,
        sign_alg=None,
        **kwargs,
    ):
        """
        Doing the second phase of the ECP conversation, the conversation
        with the IdP happens.

        :param authn_request: The AuthenticationRequest
        :param rc_url: The assertion consumer service url of the SP
        :param idp_entity_id: The EntityID of the IdP
        :param headers: Possible extra headers
        :param sign: If the message should be signed
        :return: The response from the IdP
        """

        _, destination = self.pick_binding(
            "single_sign_on_service", [BINDING_SOAP], "idpsso", entity_id=idp_entity_id
        )

        ht_args = self.apply_binding(
            BINDING_SOAP, authn_request, destination, sign=sign, sigalg=sign_alg
        )

        if headers:
            ht_args["headers"].extend(headers)

        logger.debug("[P2] Sending request: %s", ht_args["data"])

        # POST the request to the IdP
        response = self.send(**ht_args)

        logger.debug("[P2] Got IdP response: %s", response)

        if response.status_code != 200:
            raise SAMLError(
                "Request to IdP failed ({status}): {text}".format(
                    status=response.status_code, text=response.text
                )
            )

        # SAMLP response in a SOAP envelope body, ecp response in headers
        respdict = self.parse_soap_message(response.text)

        if respdict is None:
            raise SAMLError("Unexpected reply from the IdP")

        logger.debug("[P2] IdP response dict: %s", respdict)

        idp_response = respdict["body"]

        expected_tag = "Response"
        if idp_response.c_tag != expected_tag:
            raise ValueError(
                "Invalid Response tag '{invalid}' should be '{valid}'".format(
                    invalid=idp_response.c_tag, valid=expected_tag
                )
            )

        logger.debug("[P2] IdP AUTHN response: %s", idp_response)

        _ecp_response = None
        for item in respdict["header"]:
            if item.c_tag == "Response" and item.c_namespace == ecp.NAMESPACE:
                _ecp_response = item

        _acs_url = _ecp_response.assertion_consumer_service_url
        if rc_url != _acs_url:
            error = ("response_consumer_url '%s' does not match" % rc_url,
                     "assertion_consumer_service_url '%s" % _acs_url)
            # Send an error message to the SP
            _ = self.send(rc_url, "POST", data=soap.soap_fault(error))
            # Raise an exception so the user knows something went wrong
            raise SAMLError(error)

        return idp_response

    @staticmethod
    def parse_sp_ecp_response(respdict):
        if respdict is None:
            raise SAMLError("Unexpected reply from the SP")

        logger.debug("[P1] SP response dict: %s", respdict)

        # AuthnRequest in the body or not
        authn_request = respdict["body"]

        expected_tag = "AuthnRequest"
        if authn_request.c_tag != expected_tag:
            raise ValueError(
                "Invalid AuthnRequest tag '{invalid}' should be '{valid}'".format(
                    invalid=authn_request.c_tag, valid=expected_tag
                )
            )

        # ecp.RelayState among headers
        _relay_state = None
        _paos_request = None
        for item in respdict["header"]:
            if item.c_tag == "RelayState" and item.c_namespace == ecp.NAMESPACE:
                _relay_state = item
            if item.c_tag == "Request" and item.c_namespace == paos.NAMESPACE:
                _paos_request = item

        if _paos_request is None:
            raise BadRequest("Missing request")

        _rc_url = _paos_request.response_consumer_url

        return {
            "authn_request": authn_request,
            "rc_url": _rc_url,
            "relay_state": _relay_state,
        }

    def ecp_conversation(self, respdict, idp_entity_id=None):
        """

        :param respdict:
        :param idp_entity_id:
        :return:
        """

        args = self.parse_sp_ecp_response(respdict)

        # **********************
        # Phase 2 - talk to the IdP
        # **********************

        idp_response = self.phase2(idp_entity_id=idp_entity_id, **args)

        # **********************************
        # Phase 3 - back to the SP
        # **********************************

        ht_args = self.use_soap(idp_response, args["rc_url"], [args["relay_state"]])
        ht_args["headers"][0] = ('Content-Type', MIME_PAOS)
        logger.debug("[P3] Post to SP: %s", ht_args["data"])

        # POST the package from the IdP to the SP
        response = self.send(**ht_args)

        if response.status_code == 302:
            # ignore where the SP is redirecting us to and go for the
            # url I started off with.
            pass
        else:
            raise SAMLError("Error POSTing package to SP: %s" % response.text)

        logger.debug("[P3] SP response: %s", response.text)

        self.done_ecp = True
        logger.debug("Done ECP")

        return None

    @staticmethod
    def add_paos_headers(headers=None):
        if headers:
            headers = set_list2dict(headers)
            headers["PAOS"] = PAOS_HEADER_INFO
            if "Accept" in headers:
                headers["Accept"] += ";%s" % MIME_PAOS
            elif "accept" in headers:
                headers["Accept"] = headers["accept"]
                headers["Accept"] += ";%s" % MIME_PAOS
                del headers["accept"]
            headers = dict2set_list(headers)
        else:
            headers = [
                ('Accept', 'text/html; %s' % MIME_PAOS),
                ('PAOS', PAOS_HEADER_INFO)
            ]

        return headers

    def operation(self, url, idp_entity_id, op, **opargs):
        """
        This is the method that should be used by someone that wants
        to authenticate using SAML ECP

        :param url: The page that access is sought for
        :param idp_entity_id: The entity ID of the IdP that should be
            used for authentication
        :param op: Which HTTP operation (GET/POST/PUT/DELETE)
        :param opargs: Arguments to the HTTP call
        :return: The page
        """
        sp_url = self._sp

        # ********************************************
        # Phase 1 - First conversation with the SP
        # ********************************************
        # headers needed to indicate to the SP that I'm ECP enabled

        opargs["headers"] = self.add_paos_headers(opargs["headers"])
        response = self.send(sp_url, op, **opargs)
        logger.debug("[Op] SP response: %s" % response)
        print(response.text)

        if response.status_code != 200:
            raise SAMLError(
                "Request to SP failed: %s" % response.text)

        # The response might be a AuthnRequest instance in a SOAP envelope
        # body. If so it's the start of the ECP conversation
        # Two SOAP header blocks; paos:Request and ecp:Request
        # may also contain a ecp:RelayState SOAP header block
        # If channel-binding was part of the PAOS header any number of
        # <cb:ChannelBindings> header blocks may also be present
        # if 'holder-of-key' option then one or more <ecp:SubjectConfirmation>
        # header blocks may also be present
        try:
            respdict = self.parse_soap_message(response.text)
            self.ecp_conversation(respdict, idp_entity_id)

            # should by now be authenticated so this should go smoothly
            response = self.send(url, op, **opargs)
        except (soap.XmlParseError, AssertionError, KeyError):
            raise

        if response.status_code >= 400:
            raise SAMLError("Error performing operation: %s" % (
                response.text,))

        return response

    # different HTTP operations
    def delete(self, url=None, idp_entity_id=None):
        return self.operation(url, idp_entity_id, "DELETE")

    def get(self, url=None, idp_entity_id=None, headers=None):
        return self.operation(url, idp_entity_id, "GET", headers=headers)

    def post(self, url=None, data="", idp_entity_id=None, headers=None):
        return self.operation(url, idp_entity_id, "POST", data=data,
                              headers=headers)

    def put(self, url=None, data="", idp_entity_id=None, headers=None):
        return self.operation(url, idp_entity_id, "PUT", data=data,
                              headers=headers)



CLAIMS = 'http://schemas.xmlsoap.org/claims/'


MAP = {
    "identifier": "urn:oasis:names:tc:SAML:2.0:attrname-format:unspecified",
    'fro': {
        CLAIMS+'commonname': 'commonName',
        CLAIMS+'emailaddress': 'emailAddress',
        CLAIMS+'group': 'group',
        CLAIMS+'upn': 'upn',
    },
    'to': {
        'commonName': CLAIMS+'commonname',
        'emailAddress': CLAIMS+'emailaddress',
        'group': CLAIMS+'group',
        'upn': CLAIMS+'upn',
    }
}



CLAIMS = 'http://schemas.xmlsoap.org/claims/'
COM_WS_CLAIMS = 'http://schemas.xmlsoap.com/ws/2005/05/identity/claims/'
MS_CLAIMS = 'http://schemas.microsoft.com/ws/2008/06/identity/claims/'
ORG_WS_CLAIMS = 'http://schemas.xmlsoap.org/ws/2005/05/identity/claims/'


MAP = {
    "identifier": "urn:oasis:names:tc:SAML:2.0:attrname-format:unspecified",
    'fro': {
        CLAIMS+'commonname': 'commonName',
        CLAIMS+'group': 'group',
        COM_WS_CLAIMS+'denyonlysid': 'denyOnlySid',
        MS_CLAIMS+'authenticationmethod': 'authenticationMethod',
        MS_CLAIMS+'denyonlyprimarygroupsid': 'denyOnlyPrimaryGroupSid',
        MS_CLAIMS+'denyonlyprimarysid': 'denyOnlyPrimarySid',
        MS_CLAIMS+'groupsid': 'groupSid',
        MS_CLAIMS+'primarygroupsid': 'primaryGroupSid',
        MS_CLAIMS+'primarysid': 'primarySid',
        MS_CLAIMS+'role': 'role',
        MS_CLAIMS+'windowsaccountname': 'windowsAccountName',
        ORG_WS_CLAIMS+'emailaddress': 'emailAddress',
        ORG_WS_CLAIMS+'givenname': 'givenName',
        ORG_WS_CLAIMS+'name': 'name',
        ORG_WS_CLAIMS+'nameidentifier': 'nameId',
        ORG_WS_CLAIMS+'privatepersonalidentifier': 'privatePersonalId',
        ORG_WS_CLAIMS+'surname': 'surname',
        ORG_WS_CLAIMS+'upn': 'upn',
    },
    'to': {
        'authenticationMethod': MS_CLAIMS+'authenticationmethod',
        'commonName': CLAIMS+'commonname',
        'denyOnlyPrimaryGroupSid': MS_CLAIMS+'denyonlyprimarygroupsid',
        'denyOnlyPrimarySid': MS_CLAIMS+'denyonlyprimarysid',
        'denyOnlySid': COM_WS_CLAIMS+'denyonlysid',
        'emailAddress': ORG_WS_CLAIMS+'emailaddress',
        'givenName': ORG_WS_CLAIMS+'givenname',
        'group': CLAIMS+'group',
        'groupSid': MS_CLAIMS+'groupsid',
        'name': ORG_WS_CLAIMS+'name',
        'nameId': ORG_WS_CLAIMS+'nameidentifier',
        'primaryGroupSid': MS_CLAIMS+'primarygroupsid',
        'primarySid': MS_CLAIMS+'primarysid',
        'privatePersonalId': ORG_WS_CLAIMS+'privatepersonalidentifier',
        'role': MS_CLAIMS+'role',
        'surname': ORG_WS_CLAIMS+'surname',
        'upn': ORG_WS_CLAIMS+'upn',
        'windowsAccountName': MS_CLAIMS+'windowsaccountname',
    }
}



#!/usr/bin/env python
import logging

import memcache
from saml2 import time_util
from saml2.cache import TooOld, CacheError

# The assumption is that any subject may consist of data
# gathered from several different sources, all with their own
# timeout time.

logger = logging.getLogger(__name__)

def _key(prefix, name):
    return "%s_%s" % (prefix, name)

class Cache(object):
    def __init__(self, servers, debug=0):
        self._cache = memcache.Client(servers, debug)

    def delete(self, subject_id):
        entities = self.entities(subject_id)
        if entities:
            for entity_id in entities:
                if not self._cache.delete(_key(subject_id, entity_id)):
                    raise CacheError("Delete failed")

        if not self._cache.delete(subject_id):
            raise CacheError("Delete failed")

        subjects = self._cache.get("subjects")
        if subjects and subject_id in subjects:
            subjects.remove(subject_id)
            if not self._cache.set("subjects", subjects):
                raise CacheError("Set operation failed")

    def get_identity(self, subject_id, entities=None):
        """ Get all the identity information that has been received and
        are still valid about the subject.

        :param subject_id: The identifier of the subject
        :param entities: The identifiers of the entities whoes assertions are
            interesting. If the list is empty all entities are interesting.
        :return: A 2-tuple consisting of the identity information (a
            dictionary of attributes and values) and the list of entities
            whoes information has timed out.
        """
        if not entities:
            entities = self.entities(subject_id)
            if not entities:
                return {}, []

        res = {}
        oldees = []
        for (entity_id, item) in self._cache.get_multi(entities,
                                                    subject_id+'_').items():
            try:
                info = self.get_info(item)
            except TooOld:
                oldees.append(entity_id)
                continue
            for key, vals in info["ava"].items():
                try:
                    tmp = set(res[key]).union(set(vals))
                    res[key] = list(tmp)
                except KeyError:
                    res[key] = vals
        return res, oldees

    def get_info(self, item, check_not_on_or_after=True):
        """ Get session information about a subject gotten from a
        specified IdP/AA.

        :param item: Information stored
        :return: The session information as a dictionary
        """
        try:
            (timestamp, info) = item
        except ValueError:
            raise TooOld()

        if check_not_on_or_after and not time_util.not_on_or_after(timestamp):
            raise TooOld()

        return info or None

    def get(self, subject_id, entity_id, check_not_on_or_after=True):
        res = self._cache.get(_key(subject_id, entity_id))
        if not res:
            return {}
        else:
            return self.get_info(res)

    def set(self, subject_id, entity_id, info, timestamp=0):
        """ Stores session information in the cache. Assumes that the subject_id
        is unique within the context of the Service Provider.

        :param subject_id: The subject identifier
        :param entity_id: The identifier of the entity_id/receiver of an
            assertion
        :param info: The session info, the assertion is part of this
        :param timestamp: A time after which the assertion is not valid.
        """
        entities = self._cache.get(subject_id)
        if not entities:
            entities = []
            subjects = self._cache.get("subjects")
            if not subjects:
                subjects = []
            if subject_id not in subjects:
                subjects.append(subject_id)
                if not self._cache.set("subjects", subjects):
                    raise CacheError("set failed")

        if entity_id not in entities:
            entities.append(entity_id)
            if not self._cache.set(subject_id, entities):
                raise CacheError("set failed")

        # Should use memcache's expire
        if not self._cache.set(_key(subject_id, entity_id), (timestamp, info)):
            raise CacheError("set failed")

    def reset(self, subject_id, entity_id):
        """ Scrap the assertions received from a IdP or an AA about a special
        subject.

        :param subject_id: The subjects identifier
        :param entity_id: The identifier of the entity_id of the assertion
        :return:
        """
        if not self._cache.set(_key(subject_id, entity_id), {}, 0):
            raise CacheError("reset failed")

    def entities(self, subject_id):
        """ Returns all the entities of assertions for a subject, disregarding
        whether the assertion still is valid or not.

        :param subject_id: The identifier of the subject
        :return: A possibly empty list of entity identifiers
        """
        res = self._cache.get(subject_id)
        if not res:
            raise KeyError("No such subject")
        else:
            return res

    def receivers(self, subject_id):
        """ Another name for entities() just to make it more logic in the IdP
            scenario """
        return self.entities(subject_id)

    def active(self, subject_id, entity_id):
        """ Returns the status of assertions from a specific entity_id.

        :param subject_id: The ID of the subject
        :param entity_id: The entity ID of the entity_id of the assertion
        :return: True or False depending on if the assertion is still
            valid or not.
        """
        try:
            (timestamp, info) = self._cache.get(_key(subject_id, entity_id))
        except ValueError:
            return False
        except TypeError:
            return False

        # if not info:
        #     return False

        try:
            return time_util.not_on_or_after(timestamp)
        except TooOld:
            return False

    def subjects(self):
        """ Return identifiers for all the subjects that are in the cache.

        :return: list of subject identifiers
        """
        return self._cache.get("subjects")

    def update(self, subject_id, entity_id, ava):
        res = self._cache.get(_key(subject_id, entity_id))
        if res is None:
            raise KeyError("No such subject")
        else:
            info = self.get_info(res)
            if info:
                info.update(ava)
                self.set(subject_id, entity_id, info, res[0])

    def valid_to(self, subject_id, entity_id, newtime):
        try:
            (timestamp, info) = self._cache.get(_key(subject_id, entity_id))
        except ValueError:
            return False
        except TypeError:
            info = {}

        if not self._cache.set(_key(subject_id, entity_id), (newtime, info)):
            raise CacheError("valid_to failed")



#!/usr/bin/env python

#
# Generated Sun Jun 14 13:38:21 2015 by parse_xsd.py version 0.5.
#

import saml2
from saml2 import SamlBase


NAMESPACE = 'http://www.w3.org/2005/08/addressing'

class ReferenceParametersType_(SamlBase):
    """The http://www.w3.org/2005/08/addressing:ReferenceParametersType element """

    c_tag = 'ReferenceParametersType'
    c_namespace = NAMESPACE
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()

def reference_parameters_type__from_string(xml_string):
    return saml2.create_class_from_xml_string(ReferenceParametersType_, xml_string)


class MetadataType_(SamlBase):
    """The http://www.w3.org/2005/08/addressing:MetadataType element """

    c_tag = 'MetadataType'
    c_namespace = NAMESPACE
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()

def metadata_type__from_string(xml_string):
    return saml2.create_class_from_xml_string(MetadataType_, xml_string)


class RelationshipTypeOpenEnum_(SamlBase):
    """The http://www.w3.org/2005/08/addressing:RelationshipTypeOpenEnum element """

    c_tag = 'RelationshipTypeOpenEnum'
    c_namespace = NAMESPACE
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()

def relationship_type_open_enum__from_string(xml_string):
    return saml2.create_class_from_xml_string(RelationshipTypeOpenEnum_, xml_string)

class RelationshipType_(SamlBase):
    """The http://www.w3.org/2005/08/addressing:RelationshipType element """

    c_tag = 'RelationshipType'
    c_namespace = NAMESPACE
    c_value_type = {'base': 'xs:anyURI', 'enumeration': ['http://www.w3.org/2005/08/addressing/reply']}
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()

def relationship_type__from_string(xml_string):
    return saml2.create_class_from_xml_string(RelationshipType_, xml_string)


class AttributedURIType_(SamlBase):
    """The http://www.w3.org/2005/08/addressing:AttributedURIType element """

    c_tag = 'AttributedURIType'
    c_namespace = NAMESPACE
    c_value_type = {'base': 'anyURI'}
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()

def attributed_uri_type__from_string(xml_string):
    return saml2.create_class_from_xml_string(AttributedURIType_, xml_string)


class FaultCodesOpenEnumType_(SamlBase):
    """The http://www.w3.org/2005/08/addressing:FaultCodesOpenEnumType element """

    c_tag = 'FaultCodesOpenEnumType'
    c_namespace = NAMESPACE
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()

def fault_codes_open_enum_type__from_string(xml_string):
    return saml2.create_class_from_xml_string(FaultCodesOpenEnumType_, xml_string)


class FaultCodesType_(SamlBase):
    """The http://www.w3.org/2005/08/addressing:FaultCodesType element """

    c_tag = 'FaultCodesType'
    c_namespace = NAMESPACE
    c_value_type = {'base': 'xs:QName', 'enumeration': ['tns:InvalidAddressingHeader', 'tns:InvalidAddress', 'tns:InvalidEPR', 'tns:InvalidCardinality', 'tns:MissingAddressInEPR', 'tns:DuplicateMessageID', 'tns:ActionMismatch', 'tns:MessageAddressingHeaderRequired', 'tns:DestinationUnreachable', 'tns:ActionNotSupported', 'tns:EndpointUnavailable']}
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()

def fault_codes_type__from_string(xml_string):
    return saml2.create_class_from_xml_string(FaultCodesType_, xml_string)


class AttributedUnsignedLongType_(SamlBase):
    """The http://www.w3.org/2005/08/addressing:AttributedUnsignedLongType element """

    c_tag = 'AttributedUnsignedLongType'
    c_namespace = NAMESPACE
    c_value_type = {'base': 'unsignedLong'}
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()

def attributed_unsigned_long_type__from_string(xml_string):
    return saml2.create_class_from_xml_string(AttributedUnsignedLongType_, xml_string)


class AttributedQNameType_(SamlBase):
    """The http://www.w3.org/2005/08/addressing:AttributedQNameType element """

    c_tag = 'AttributedQNameType'
    c_namespace = NAMESPACE
    c_value_type = {'base': 'QName'}
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()

def attributed_q_name_type__from_string(xml_string):
    return saml2.create_class_from_xml_string(AttributedQNameType_, xml_string)


class ProblemIRI(AttributedURIType_):
    """The http://www.w3.org/2005/08/addressing:ProblemIRI element """

    c_tag = 'ProblemIRI'
    c_namespace = NAMESPACE
    c_children = AttributedURIType_.c_children.copy()
    c_attributes = AttributedURIType_.c_attributes.copy()
    c_child_order = AttributedURIType_.c_child_order[:]
    c_cardinality = AttributedURIType_.c_cardinality.copy()

def problem_iri_from_string(xml_string):
    return saml2.create_class_from_xml_string(ProblemIRI, xml_string)


class EndpointReferenceType_Address(AttributedURIType_):

    c_tag = 'Address'
    c_namespace = NAMESPACE
    c_children = AttributedURIType_.c_children.copy()
    c_attributes = AttributedURIType_.c_attributes.copy()
    c_child_order = AttributedURIType_.c_child_order[:]
    c_cardinality = AttributedURIType_.c_cardinality.copy()

def endpoint_reference_type__address_from_string(xml_string):
    return saml2.create_class_from_xml_string(EndpointReferenceType_Address, xml_string)


class ReferenceParameters(ReferenceParametersType_):
    """The http://www.w3.org/2005/08/addressing:ReferenceParameters element """

    c_tag = 'ReferenceParameters'
    c_namespace = NAMESPACE
    c_children = ReferenceParametersType_.c_children.copy()
    c_attributes = ReferenceParametersType_.c_attributes.copy()
    c_child_order = ReferenceParametersType_.c_child_order[:]
    c_cardinality = ReferenceParametersType_.c_cardinality.copy()

def reference_parameters_from_string(xml_string):
    return saml2.create_class_from_xml_string(ReferenceParameters, xml_string)


class Metadata(MetadataType_):
    """The http://www.w3.org/2005/08/addressing:Metadata element """

    c_tag = 'Metadata'
    c_namespace = NAMESPACE
    c_children = MetadataType_.c_children.copy()
    c_attributes = MetadataType_.c_attributes.copy()
    c_child_order = MetadataType_.c_child_order[:]
    c_cardinality = MetadataType_.c_cardinality.copy()

def metadata_from_string(xml_string):
    return saml2.create_class_from_xml_string(Metadata, xml_string)


class MessageID(AttributedURIType_):
    """The http://www.w3.org/2005/08/addressing:MessageID element """

    c_tag = 'MessageID'
    c_namespace = NAMESPACE
    c_children = AttributedURIType_.c_children.copy()
    c_attributes = AttributedURIType_.c_attributes.copy()
    c_child_order = AttributedURIType_.c_child_order[:]
    c_cardinality = AttributedURIType_.c_cardinality.copy()

def message_id_from_string(xml_string):
    return saml2.create_class_from_xml_string(MessageID, xml_string)


class RelatesToType_(SamlBase):
    """The http://www.w3.org/2005/08/addressing:RelatesToType element """

    c_tag = 'RelatesToType'
    c_namespace = NAMESPACE
    c_value_type = {'base': 'anyURI'}
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()
    c_attributes['RelationshipType'] = ('relationship_type', RelationshipTypeOpenEnum_, False)

    def __init__(self,
            relationship_type='http://www.w3.org/2005/08/addressing/reply',
            text=None,
            extension_elements=None,
            extension_attributes=None,
        ):
        SamlBase.__init__(self, 
                text=text,
                extension_elements=extension_elements,
                extension_attributes=extension_attributes,
                )
        self.relationship_type=relationship_type

def relates_to_type__from_string(xml_string):
    return saml2.create_class_from_xml_string(RelatesToType_, xml_string)


class To(AttributedURIType_):
    """The http://www.w3.org/2005/08/addressing:To element """

    c_tag = 'To'
    c_namespace = NAMESPACE
    c_children = AttributedURIType_.c_children.copy()
    c_attributes = AttributedURIType_.c_attributes.copy()
    c_child_order = AttributedURIType_.c_child_order[:]
    c_cardinality = AttributedURIType_.c_cardinality.copy()

def to_from_string(xml_string):
    return saml2.create_class_from_xml_string(To, xml_string)


class Action(AttributedURIType_):
    """The http://www.w3.org/2005/08/addressing:Action element """

    c_tag = 'Action'
    c_namespace = NAMESPACE
    c_children = AttributedURIType_.c_children.copy()
    c_attributes = AttributedURIType_.c_attributes.copy()
    c_child_order = AttributedURIType_.c_child_order[:]
    c_cardinality = AttributedURIType_.c_cardinality.copy()

def action_from_string(xml_string):
    return saml2.create_class_from_xml_string(Action, xml_string)


class RetryAfter(AttributedUnsignedLongType_):
    """The http://www.w3.org/2005/08/addressing:RetryAfter element """

    c_tag = 'RetryAfter'
    c_namespace = NAMESPACE
    c_children = AttributedUnsignedLongType_.c_children.copy()
    c_attributes = AttributedUnsignedLongType_.c_attributes.copy()
    c_child_order = AttributedUnsignedLongType_.c_child_order[:]
    c_cardinality = AttributedUnsignedLongType_.c_cardinality.copy()

def retry_after_from_string(xml_string):
    return saml2.create_class_from_xml_string(RetryAfter, xml_string)


class ProblemHeaderQName(AttributedQNameType_):
    """The http://www.w3.org/2005/08/addressing:ProblemHeaderQName element """

    c_tag = 'ProblemHeaderQName'
    c_namespace = NAMESPACE
    c_children = AttributedQNameType_.c_children.copy()
    c_attributes = AttributedQNameType_.c_attributes.copy()
    c_child_order = AttributedQNameType_.c_child_order[:]
    c_cardinality = AttributedQNameType_.c_cardinality.copy()

def problem_header_q_name_from_string(xml_string):
    return saml2.create_class_from_xml_string(ProblemHeaderQName, xml_string)


class ProblemActionType_SoapAction(SamlBase):

    c_tag = 'SoapAction'
    c_namespace = NAMESPACE
    c_value_type = {'base': 'anyURI'}
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()

def problem_action_type__soap_action_from_string(xml_string):
    return saml2.create_class_from_xml_string(ProblemActionType_SoapAction, xml_string)


class ProblemActionType_(SamlBase):
    """The http://www.w3.org/2005/08/addressing:ProblemActionType element """

    c_tag = 'ProblemActionType'
    c_namespace = NAMESPACE
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()
    c_children['{http://www.w3.org/2005/08/addressing}Action'] = ('action', Action)
    c_cardinality['action'] = {"min":0, "max":1}
    c_children['{http://www.w3.org/2005/08/addressing}SoapAction'] = ('soap_action', ProblemActionType_SoapAction)
    c_cardinality['soap_action'] = {"min":0, "max":1}
    c_child_order.extend(['action', 'soap_action'])

    def __init__(self,
            action=None,
            soap_action=None,
            text=None,
            extension_elements=None,
            extension_attributes=None,
        ):
        SamlBase.__init__(self, 
                text=text,
                extension_elements=extension_elements,
                extension_attributes=extension_attributes,
                )
        self.action=action
        self.soap_action=soap_action

def problem_action_type__from_string(xml_string):
    return saml2.create_class_from_xml_string(ProblemActionType_, xml_string)


class EndpointReferenceType_(SamlBase):
    """The http://www.w3.org/2005/08/addressing:EndpointReferenceType element """

    c_tag = 'EndpointReferenceType'
    c_namespace = NAMESPACE
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()
    c_children['{http://www.w3.org/2005/08/addressing}Address'] = ('address', EndpointReferenceType_Address)
    c_children['{http://www.w3.org/2005/08/addressing}ReferenceParameters'] = ('reference_parameters', ReferenceParameters)
    c_cardinality['reference_parameters'] = {"min":0, "max":1}
    c_children['{http://www.w3.org/2005/08/addressing}Metadata'] = ('metadata', Metadata)
    c_cardinality['metadata'] = {"min":0, "max":1}
    c_child_order.extend(['address', 'reference_parameters', 'metadata'])

    def __init__(self,
            address=None,
            reference_parameters=None,
            metadata=None,
            text=None,
            extension_elements=None,
            extension_attributes=None,
        ):
        SamlBase.__init__(self, 
                text=text,
                extension_elements=extension_elements,
                extension_attributes=extension_attributes,
                )
        self.address=address
        self.reference_parameters=reference_parameters
        self.metadata=metadata

def endpoint_reference_type__from_string(xml_string):
    return saml2.create_class_from_xml_string(EndpointReferenceType_, xml_string)


class RelatesTo(RelatesToType_):
    """The http://www.w3.org/2005/08/addressing:RelatesTo element """

    c_tag = 'RelatesTo'
    c_namespace = NAMESPACE
    c_children = RelatesToType_.c_children.copy()
    c_attributes = RelatesToType_.c_attributes.copy()
    c_child_order = RelatesToType_.c_child_order[:]
    c_cardinality = RelatesToType_.c_cardinality.copy()

def relates_to_from_string(xml_string):
    return saml2.create_class_from_xml_string(RelatesTo, xml_string)


class ReplyTo(EndpointReferenceType_):
    """The http://www.w3.org/2005/08/addressing:ReplyTo element """

    c_tag = 'ReplyTo'
    c_namespace = NAMESPACE
    c_children = EndpointReferenceType_.c_children.copy()
    c_attributes = EndpointReferenceType_.c_attributes.copy()
    c_child_order = EndpointReferenceType_.c_child_order[:]
    c_cardinality = EndpointReferenceType_.c_cardinality.copy()

def reply_to_from_string(xml_string):
    return saml2.create_class_from_xml_string(ReplyTo, xml_string)


class From(EndpointReferenceType_):
    """The http://www.w3.org/2005/08/addressing:From element """

    c_tag = 'From'
    c_namespace = NAMESPACE
    c_children = EndpointReferenceType_.c_children.copy()
    c_attributes = EndpointReferenceType_.c_attributes.copy()
    c_child_order = EndpointReferenceType_.c_child_order[:]
    c_cardinality = EndpointReferenceType_.c_cardinality.copy()

def from_from_string(xml_string):
    return saml2.create_class_from_xml_string(From, xml_string)


class FaultTo(EndpointReferenceType_):
    """The http://www.w3.org/2005/08/addressing:FaultTo element """

    c_tag = 'FaultTo'
    c_namespace = NAMESPACE
    c_children = EndpointReferenceType_.c_children.copy()
    c_attributes = EndpointReferenceType_.c_attributes.copy()
    c_child_order = EndpointReferenceType_.c_child_order[:]
    c_cardinality = EndpointReferenceType_.c_cardinality.copy()

def fault_to_from_string(xml_string):
    return saml2.create_class_from_xml_string(FaultTo, xml_string)


class ProblemAction(ProblemActionType_):
    """The http://www.w3.org/2005/08/addressing:ProblemAction element """

    c_tag = 'ProblemAction'
    c_namespace = NAMESPACE
    c_children = ProblemActionType_.c_children.copy()
    c_attributes = ProblemActionType_.c_attributes.copy()
    c_child_order = ProblemActionType_.c_child_order[:]
    c_cardinality = ProblemActionType_.c_cardinality.copy()

def problem_action_from_string(xml_string):
    return saml2.create_class_from_xml_string(ProblemAction, xml_string)


class EndpointReference(EndpointReferenceType_):
    """The http://www.w3.org/2005/08/addressing:EndpointReference element """

    c_tag = 'EndpointReference'
    c_namespace = NAMESPACE
    c_children = EndpointReferenceType_.c_children.copy()
    c_attributes = EndpointReferenceType_.c_attributes.copy()
    c_child_order = EndpointReferenceType_.c_child_order[:]
    c_cardinality = EndpointReferenceType_.c_cardinality.copy()

def endpoint_reference_from_string(xml_string):
    return saml2.create_class_from_xml_string(EndpointReference, xml_string)


ELEMENT_FROM_STRING = {
    EndpointReference.c_tag: endpoint_reference_from_string,
    EndpointReferenceType_.c_tag: endpoint_reference_type__from_string,
    ReferenceParameters.c_tag: reference_parameters_from_string,
    ReferenceParametersType_.c_tag: reference_parameters_type__from_string,
    Metadata.c_tag: metadata_from_string,
    MetadataType_.c_tag: metadata_type__from_string,
    MessageID.c_tag: message_id_from_string,
    RelatesTo.c_tag: relates_to_from_string,
    RelatesToType_.c_tag: relates_to_type__from_string,
    RelationshipTypeOpenEnum_.c_tag: relationship_type_open_enum__from_string,
    RelationshipType_.c_tag: relationship_type__from_string,
    ReplyTo.c_tag: reply_to_from_string,
    From.c_tag: from_from_string,
    FaultTo.c_tag: fault_to_from_string,
    To.c_tag: to_from_string,
    Action.c_tag: action_from_string,
    AttributedURIType_.c_tag: attributed_uri_type__from_string,
    FaultCodesOpenEnumType_.c_tag: fault_codes_open_enum_type__from_string,
    FaultCodesType_.c_tag: fault_codes_type__from_string,
    RetryAfter.c_tag: retry_after_from_string,
    AttributedUnsignedLongType_.c_tag: attributed_unsigned_long_type__from_string,
    ProblemHeaderQName.c_tag: problem_header_q_name_from_string,
    AttributedQNameType_.c_tag: attributed_q_name_type__from_string,
    ProblemIRI.c_tag: problem_iri_from_string,
    ProblemAction.c_tag: problem_action_from_string,
    ProblemActionType_.c_tag: problem_action_type__from_string,
    EndpointReferenceType_Address.c_tag: endpoint_reference_type__address_from_string,
    ProblemActionType_SoapAction.c_tag: problem_action_type__soap_action_from_string,
}

ELEMENT_BY_TAG = {
    'EndpointReference': EndpointReference,
    'EndpointReferenceType': EndpointReferenceType_,
    'ReferenceParameters': ReferenceParameters,
    'ReferenceParametersType': ReferenceParametersType_,
    'Metadata': Metadata,
    'MetadataType': MetadataType_,
    'MessageID': MessageID,
    'RelatesTo': RelatesTo,
    'RelatesToType': RelatesToType_,
    'RelationshipTypeOpenEnum': RelationshipTypeOpenEnum_,
    'RelationshipType': RelationshipType_,
    'ReplyTo': ReplyTo,
    'From': From,
    'FaultTo': FaultTo,
    'To': To,
    'Action': Action,
    'AttributedURIType': AttributedURIType_,
    'FaultCodesOpenEnumType': FaultCodesOpenEnumType_,
    'FaultCodesType': FaultCodesType_,
    'RetryAfter': RetryAfter,
    'AttributedUnsignedLongType': AttributedUnsignedLongType_,
    'ProblemHeaderQName': ProblemHeaderQName,
    'AttributedQNameType': AttributedQNameType_,
    'ProblemIRI': ProblemIRI,
    'ProblemAction': ProblemAction,
    'ProblemActionType': ProblemActionType_,
    'Address': EndpointReferenceType_Address,
    'SoapAction': ProblemActionType_SoapAction,
}


def factory(tag, **kwargs):
    return ELEMENT_BY_TAG[tag](**kwargs)




import calendar
import six
from six.moves import http_cookiejar
import copy
import re
from six.moves.urllib.parse import urlparse
from six.moves.urllib.parse import urlencode
import requests
import time
from six.moves.http_cookies import SimpleCookie
from saml2.time_util import utc_now
from saml2 import class_name, SAMLError
from saml2.pack import http_post_message
from saml2.pack import make_soap_enveloped_saml_thingy

import logging

logger = logging.getLogger(__name__)

if requests.__version__ < "2.0.0":
    DICT_HEADERS = False
else:
    DICT_HEADERS = True

__author__ = 'rolandh'

ATTRS = {"version": None,
         "name": "",
         "value": None,
         "port": None,
         "port_specified": False,
         "domain": "",
         "domain_specified": False,
         "domain_initial_dot": False,
         "path": "",
         "path_specified": False,
         "secure": False,
         "expires": None,
         "discard": True,
         "comment": None,
         "comment_url": None,
         "rest": "",
         "rfc2109": True}

PAIRS = {
    "port": "port_specified",
    "domain": "domain_specified",
    "path": "path_specified"
}


class ConnectionError(SAMLError):
    pass


class HTTPError(SAMLError):
    pass


TIME_FORMAT = ["%d-%b-%Y %H:%M:%S %Z", "%d-%b-%y %H:%M:%S %Z",
               "%d %b %Y %H:%M:%S %Z"]


def _since_epoch(cdate):
    """
    :param cdate: date format 'Wed, 06-Jun-2012 01:34:34 GMT'
    :return: UTC time
    """

    if len(cdate) < 29:  # somethings broken
        if len(cdate) < 5:
            return utc_now()

    cdate = cdate[5:] # assume short weekday, i.e. do not support obsolete RFC 1036 date format
    t = -1
    for time_format in TIME_FORMAT :
        try:
            t = time.strptime(cdate, time_format)   # e.g. 18-Apr-2014 12:30:51 GMT
        except ValueError:
            pass
        else:
            break

    if t == -1:
        err = 'ValueError: Date "{0}" does not match any of: {1}'.format(
            cdate, TIME_FORMAT
        )
        raise Exception(err)

    return calendar.timegm(t)


def set_list2dict(sl):
    return dict(sl)


def dict2set_list(dic):
    return [(k, v) for k, v in dic.items()]


class HTTPBase(object):
    def __init__(self, verify=True, ca_bundle=None, key_file=None,
                 cert_file=None):
        self.request_args = {"allow_redirects": False}
        #self.cookies = {}
        self.cookiejar = http_cookiejar.CookieJar()

        self.request_args["verify"] = verify
        if verify:
            if ca_bundle:
                self.request_args["verify"] = ca_bundle
            if key_file:
                self.request_args["cert"] = (cert_file, key_file)

        self.sec = None
        self.user = None
        self.passwd = None

    def cookies(self, url):
        """
        Return cookies that are matching the path and are still valid

        :param url:
        :return:
        """
        part = urlparse(url)

        #if part.port:
        #    _domain = "%s:%s" % (part.hostname, part.port)
        #else:
        _domain = part.hostname

        cookie_dict = {}
        now = utc_now()
        for _, a in list(self.cookiejar._cookies.items()):
            for _, b in a.items():
                for cookie in list(b.values()):
                    # print(cookie)
                    if cookie.expires and cookie.expires <= now:
                        continue
                    if not re.search("%s$" % cookie.domain, _domain):
                        continue
                    if not re.match(cookie.path, part.path):
                        continue

                    cookie_dict[cookie.name] = cookie.value

        return cookie_dict

    def set_cookie(self, kaka, request):
        """Returns a http_cookiejar.Cookie based on a set-cookie header line"""

        if not kaka:
            return

        part = urlparse(request.url)
        _domain = part.hostname
        logger.debug("%s: '%s'", _domain, kaka)

        for cookie_name, morsel in kaka.items():
            std_attr = ATTRS.copy()
            std_attr["name"] = cookie_name
            _tmp = morsel.coded_value
            if _tmp.startswith('"') and _tmp.endswith('"'):
                std_attr["value"] = _tmp[1:-1]
            else:
                std_attr["value"] = _tmp

            std_attr["version"] = 0
            # copy attributes that have values
            for attr in morsel.keys():
                if attr in ATTRS:
                    if morsel[attr]:
                        if attr == "expires":
                            std_attr[attr] = _since_epoch(morsel[attr])
                        elif attr == "path":
                            if morsel[attr].endswith(","):
                                std_attr[attr] = morsel[attr][:-1]
                            else:
                                std_attr[attr] = morsel[attr]
                        else:
                            std_attr[attr] = morsel[attr]
                elif attr == "max-age":
                    if morsel["max-age"]:
                        std_attr["expires"] = time.time() + int(morsel["max-age"])

            for att, item in PAIRS.items():
                if std_attr[att]:
                    std_attr[item] = True

            if std_attr["domain"]:
                if std_attr["domain"].startswith("."):
                    std_attr["domain_initial_dot"] = True
            else:
                std_attr["domain"] = _domain
                std_attr["domain_specified"] = True

            if morsel["max-age"] == 0:
                try:
                    self.cookiejar.clear(domain=std_attr["domain"],
                                         path=std_attr["path"],
                                         name=std_attr["name"])
                except ValueError:
                    pass
            elif std_attr["expires"] and std_attr["expires"] < utc_now():
                try:
                    self.cookiejar.clear(domain=std_attr["domain"],
                                         path=std_attr["path"],
                                         name=std_attr["name"])
                except ValueError:
                    pass
            else:
                new_cookie = http_cookiejar.Cookie(**std_attr)
                self.cookiejar.set_cookie(new_cookie)

    def send(self, url, method="GET", **kwargs):
        _kwargs = copy.copy(self.request_args)
        if kwargs:
            _kwargs.update(kwargs)

        if self.cookiejar:
            _cd = self.cookies(url)
            if _cd:
                _kwargs["cookies"] = _cd

        if self.user and self.passwd:
            _kwargs["auth"] = (self.user, self.passwd)

        if "headers" in _kwargs and isinstance(_kwargs["headers"], list):
            if DICT_HEADERS:
                # requests.request wants a dict of headers, not a list of tuples
                _kwargs["headers"] = dict(_kwargs["headers"])

        try:
            logger.debug("%s to %s", method, url)
            for arg in ["cookies", "data", "auth"]:
                try:
                    logger.debug("%s: %s", arg.upper(), _kwargs[arg])
                except KeyError:
                    pass
            r = requests.request(method, url, **_kwargs)
            logger.debug("Response status: %s", r.status_code)
        except requests.ConnectionError as exc:
            raise ConnectionError("%s" % exc)

        try:
            self.set_cookie(SimpleCookie(r.headers["set-cookie"]), r)
        except AttributeError:
            pass
        except KeyError:
            pass

        return r

    @staticmethod
    def use_http_artifact(message, destination="", relay_state=""):
        if relay_state:
            query = urlencode({"SAMLart": message,
                               "RelayState": relay_state})
        else:
            query = urlencode({"SAMLart": message})
        info = {
            "data": "",
            "url": "%s?%s" % (destination, query)
        }
        return info

    @staticmethod
    def use_http_uri(message, typ, destination="", relay_state=""):
        if "\n" in message:
            data = message.split("\n")[1]
        else:
            data = message.strip()
        if typ == "SAMLResponse":
            info = {
                "data": data,
                "headers": [
                    ("Content-Type", "application/samlassertion+xml"),
                    ("Cache-Control", "no-cache, no-store"),
                    ("Pragma", "no-cache")
                ]
            }
        elif typ == "SAMLRequest":
            # msg should be an identifier
            if relay_state:
                query = urlencode({"ID": message,
                                   "RelayState": relay_state})
            else:
                query = urlencode({"ID": message})
            info = {
                "data": "",
                "url": "%s?%s" % (destination, query)
            }
        else:
            raise NotImplementedError

        return info

    def use_soap(self, request, destination="", soap_headers=None, sign=False,
                 **kwargs):
        """
        Construct the necessary information for using SOAP+POST

        :param request:
        :param destination:
        :param soap_headers:
        :param sign:
        :return: dictionary
        """
        headers = [("content-type", "application/soap+xml")]

        soap_message = make_soap_enveloped_saml_thingy(request, soap_headers)

        logger.debug("SOAP message: %s", soap_message)

        if sign and self.sec:
            _signed = self.sec.sign_statement(soap_message,
                                              class_name=class_name(request),
                                              node_id=request.id)
            soap_message = _signed

        return {"url": destination, "method": "POST",
                "data": soap_message, "headers": headers}

    def send_using_soap(self, request, destination, headers=None, sign=False):
        """
        Send a message using SOAP+POST

        :param request:
        :param destination:
        :param headers:
        :param sign:
        :return:
        """

        # _response = self.server.post(soap_message, headers, path=path)
        try:
            args = self.use_soap(request, destination, headers, sign)
            args["headers"] = dict(args["headers"])
            response = self.send(**args)
        except Exception as exc:
            logger.info("HTTPClient exception: %s", exc)
            raise

        if response.status_code == 200:
            logger.info("SOAP response: %s", response.text)
            return response
        else:
            raise HTTPError("%d:%s" % (response.status_code, response.content))

    def add_credentials(self, user, passwd):
        self.user = user
        self.passwd = passwd
